namespace WICS.Generator.MSBuildTasks
{
	using System;
	using System.Collections.Generic;
	using System.Collections;
	using System.Text;
	using System.IO;

	using Microsoft.Build.Framework;
	using Microsoft.Build.Utilities;

	/// <summary>
	/// Exposes methods to add directories to <see cref="ITaskItem"/> collection.
	/// </summary>
	public class AddDirectories: Task
	{
		private string directory;
		private bool delete = false;
		private ITaskItem[] itemGroup;
		private bool returnOnlyDirectories = false;
		private string placeholder;
		/// <summary>
		/// Deletes placeholder files in the given directory.
		/// </summary>
		public bool DeletePlaceholders
		{
			get { return delete; }
			set { delete = value; }
		}

		/// <summary>
		/// Gets or sets a value of the directory.
		/// </summary>
		public string Directory
		{
			get { return this.directory; }
			set { this.directory = value; }
		}

		/// <summary>
		/// The item group that will be cleaned and returned.
		/// </summary>
		[Output]
		public ITaskItem[] ItemGroup
		{
			get { return this.itemGroup; }
			set { this.itemGroup = value; }
		}

		/// <summary>
		/// The name of the placeholder to create, delete and search for.
		/// </summary>
		/// <remarks>The default placeholder is "file.placeholder".</remarks>
		public string Placeholder
		{
			get { return this.placeholder; }
			set { this.placeholder = value; }
		}

		/// <summary>
		/// Gets or sets a value to determine if the <c>ItemGroup</c> will have only directories listed.
		/// </summary>
		public bool ReturnOnlyDirectories
		{
			get { return this.returnOnlyDirectories; }
			set { this.returnOnlyDirectories = value; }
		}

		/// <summary>
		/// Executes the task.
		/// </summary>
		/// <returns>Returns true if the task completed sucessfully.</returns>
		public override bool Execute ( )
		{
			if ( this.placeholder == null )
				this.placeholder = "file.placeholder";

			if ( this.Directory != null )
			{
				if ( !System.IO.Directory.Exists ( this.Directory ) )
					throw new DirectoryNotFoundException ( string.Format ( "The directory must exist to be processed. {0} was not found.", this.Directory ) );

				DirectoryInfo dir = new DirectoryInfo ( this.Directory );

				if ( !string.IsNullOrEmpty ( this.Directory ) && !this.DeletePlaceholders )
					this.WalkDirectory ( dir );

				if ( this.DeletePlaceholders )
					this.WalkDirectoryDeleting ( new DirectoryInfo ( this.Directory ) );
			}

			if ( this.ItemGroup != null )
			{
				List<ITaskItem> allItems = new List<ITaskItem> ( );
				this.AppendDirectories ( allItems );
				this.ItemGroup = allItems.ToArray ( );
			}

			return !Log.HasLoggedErrors;
		}

		private void WalkDirectory ( DirectoryInfo dir )
		{
			if ( !File.Exists ( dir.FullName + @"\" + this.placeholder ) )
				File.CreateText ( dir.FullName + @"\" + this.placeholder );

			foreach ( DirectoryInfo subDir in dir.GetDirectories ( ) )
				this.WalkDirectory ( subDir );
		}

		private void WalkDirectoryDeleting ( DirectoryInfo dir )
		{
			try
			{
				if ( dir.GetFiles ( this.placeholder ).Length > 0 )
					File.Delete ( string.Format ( @"{0}\{1}", dir.FullName, this.Placeholder ) );
			}
			catch ( System.IO.IOException )
			{
				Log.LogError ( "The placeholder ({0}) in the {1} directory has not been released yet.", this.Placeholder, dir.FullName );
				Log.LogError ( "Placeholders might be getting created and deleted frequently for this directory." );
			}

			foreach ( DirectoryInfo subDir in dir.GetDirectories ( ) )
				this.WalkDirectoryDeleting ( subDir );
		}

		private void AppendDirectories ( List<ITaskItem> allItems)
		{
			foreach ( ITaskItem item in this.ItemGroup )
			{
				bool isPlaceholder = item.GetMetadata ( "Filename" ) + item.GetMetadata ( "Extension" ) == this.placeholder;
				string dirName = this.GetIdentityDirectory ( item );

				if ( isPlaceholder )
				{
					TaskItem dir = new TaskItem ( dirName );
					
					// If the directory should have additional metadata it will be set on the placeholder item.
					if ( isPlaceholder )
						item.CopyMetadataTo ( dir );

					allItems.Add ( dir );
				}

				if ( !isPlaceholder && !this.returnOnlyDirectories )
					allItems.Add ( item );
			}
		}

		private string GetIdentityDirectory ( ITaskItem item )
		{
			string id = "Identity";

			if ( item.GetMetadata ( id ).LastIndexOf ( '\\' ) != -1 )
			{
				return item.GetMetadata ( id ).Substring ( 0, item.GetMetadata ( id ).LastIndexOf ( '\\' ) );
			}

			return ".";
		}
	}
}