namespace WICS.Generator.Processors
{
	using System;
	using System.Collections.Generic;
	using System.Text;
	using System.Xml;
	using System.Runtime.InteropServices;

	using Microsoft.Win32;
	using Microsoft.Build.Framework;

	public class ComProcessor : Processor
	{
		private System.IO.FileInfo localFile;

		/// <summary>
		/// Initializes the object.
		/// </summary>
		/// <param name="file">The <see cref="ITaskItem"/> file.</param>
		/// <param name="dirId">Sets the parent directories Id.</param>
		public ComProcessor ( ITaskItem file, string dirId ) : base ( file, dirId ) { }

		/// <summary>
		/// Executes processing on the file.
		/// </summary>
		/// <param name="writer">The <see cref="XmlWriter"/> to append to.</param>
		public override void Execute ( XmlWriter writer )
		{
			IntPtr dllHandle = Common.LoadFile ( new System.IO.FileInfo ( File.GetMetadata ( "FullPath" ) ) );

			if ( dllHandle == IntPtr.Zero )
			{
				string err = string.Format ( "Error #{0}: {1}", Marshal.GetLastWin32Error ( ).ToString ( ),
					Common.GetErrorMessage ( Marshal.GetLastWin32Error ( ) ) );

				throw ( new System.IO.FileLoadException ( err, File.ItemSpec ) );
			}
			else
			{
				this.localFile = new System.IO.FileInfo ( this.File.ItemSpec );
				this.WriteFile ( writer );
				
				this.WriteComComponent ( writer );
			}
		}

		/// <summary>
		/// Writes the COM+ component in WiX format.
		/// </summary>
		/// <param name="writer">A <see cref="XmlWriter"/> to append data.</param>
		private void WriteComComponent ( XmlWriter writer )
		{
			bool success = false;

			try
			{
				Common.StartOverridingRegHives ( );

				Common.DynamicPInvoke ( File.GetMetadata ( "FullPath" ), "DllRegisterServer", typeof ( int ), null, null );

				success = true;
			}
			finally
			{
				try
				{
					Common.EndOverridingRegHives ( );
					this.WriteRegistry ( success ? writer : null );
				}
				finally
				{
					Common.RemovePrivateRegKey ( );
				}
			}
		}

		private void WriteRegistry ( XmlWriter writer )
		{
			if ( writer != null )
			{
				this.CallWriteSubKey ( writer, Common.RegistryOverrideRootKey + @"\HKCR" );
				this.CallWriteSubKey ( writer, Common.RegistryOverrideRootKey + @"\HKCU" );
				this.CallWriteSubKey ( writer, Common.RegistryOverrideRootKey + @"\HKU" );
				this.CallWriteSubKey ( writer, Common.RegistryOverrideRootKey + @"\HKU" );
			}
		}

		private void CallWriteSubKey ( XmlWriter writer, string hive )
		{
			using ( RegistryKey key = Registry.LocalMachine.OpenSubKey ( hive ) )
				this.WriteSubKeys ( writer,
					Registry.LocalMachine.ToString ( ) + @"\" + Common.RegistryOverrideRootKey + @"\",
					key );
		}

		/// <summary>
		/// Writes subkeys of this registry key to the writer.
		/// </summary>
		/// <param name="writer">Writer to output to.</param>
		/// <param name="hiveLocation">The current hive that is being interated.</param>
		/// <param name="key">Key to have sub keys written</param>
		private void WriteSubKeys ( XmlWriter writer, string hiveLocation, RegistryKey key )
		{
			string root;
			string path = key.Name.Replace ( hiveLocation, "" );
			string[] names = key.GetValueNames ( );
			string[] subkeys = key.GetSubKeyNames ( );

			int slash = path.IndexOf ( '\\' );
			if ( -1 == slash )
			{
				root = path;
				path = String.Empty;
			}
			else
			{
				root = path.Substring ( 0, slash );
				path = path.Substring ( slash + 1, path.Length - slash - 1 );
			}

			// make sure the root is valid
			switch ( root )
			{
				case "HKLM":
				case "HKCR":
				case "HKCU":
				case "HKU":
					break;
				default:
					throw new InvalidOperationException ( String.Concat ( "Unknown key: ", key.Name ) );
			}

			// registry names
			for ( int i = 0; i < names.Length; ++i )
			{
				object value = key.GetValue ( names[i] );

				this.WriteRegistryElement ( writer, root, path, names[i], value );
			}

			// do subkeys
			for ( int i = 0; i < subkeys.Length; ++i )
			{
				RegistryKey subkey = key.OpenSubKey ( subkeys[i] );
				this.WriteSubKeys ( writer, hiveLocation, subkey );
			}

			// if there were no names or subkeys, make sure the path is in
			if ( 0 < path.Length && 0 == names.Length && 0 == subkeys.Length )
			{
				this.WriteRegistryElement ( writer, root, path, null, null );
			}
		}

		/// <summary>
		/// Writes a Registry element to the writer.
		/// </summary>
		/// <param name="writer">Writer to output to.</param>
		/// <param name="root">Root for the key.</param>
		/// <param name="path">Path to the key.</param>
		/// <param name="name">Name of the key.</param>
		/// <param name="value">Value for the key.</param>
		private void WriteRegistryElement ( XmlWriter writer, string root, string path, string name, object value )
		{
			writer.WriteStartElement ( "Registry" );
			writer.WriteAttributeString ( "Id", Common.Guids.CreateGuid ( "RegistryId", GuidManager.GuidType.Id ) );
			writer.WriteAttributeString ( "Root", root );
			writer.WriteAttributeString ( "Key", path );

			if ( null != name && 0 != name.Length )
			{
				writer.WriteAttributeString ( "Name", name );
			}

			if ( null != value )
			{
				string stringValue = value as String;

				if ( null != stringValue )
				{
					if ( 0 != stringValue.Length )
					{
						writer.WriteAttributeString ( "Value", this.GetValue ( stringValue ) );
						writer.WriteAttributeString ( "Type", "string" );
					}
				}
				else
				{
					writer.WriteAttributeString ( "Value", value.ToString ( ) );

					if ( value is int || value is long || value is uint || value is ulong )
					{
						writer.WriteAttributeString ( "Type", "integer" );
					}
					else
					{
						writer.WriteAttributeString ( "Type", "unknown" );
					}
				}
			}

			writer.WriteEndElement ( );
		}

		private string GetValue ( string value )
		{
			if ( value.Contains ( this.localFile.Name ) )
				return "[#" + this.FileId + "]";

			if ( value.Contains ( this.GetShortFileName ( false ) ) )
				return "[#" + this.FileId + "]";

			if ( value.Contains ( this.GetShortFileName ( true ) ) )
				return "[#" + this.FileId + "]";

			if ( value.Contains( this.GetFileDirectory ( ) ) )
				return "[" + this.DirectoryId + "]";

			if ( value.Contains ( Environment.GetEnvironmentVariable ( "windir" ) ) )
				return this.GetWindowsDirectory ( value );

			return value;
		}

		private string GetShortFileName ( bool withTilde )
		{
			return withTilde ? Common.GenerateShortName ( this.localFile ).Replace ( '_', '~' ) : Common.GenerateShortName ( this.localFile );
		}

		private string GetFileDirectory ( )
		{
			return this.localFile.DirectoryName;
		}

		private string GetWindowsDirectory ( string value )
		{
			if ( value.Contains ( Environment.SystemDirectory ) )
				return value.Replace ( Environment.SystemDirectory, "[SystemFolder]" );

			return value.Replace ( Environment.GetEnvironmentVariable ( "windir" ), "[WindowsFolder]" );
		}
	}
}