namespace WICS.Generator.Processors
{
	using System;
	using System.Collections.Generic;
	using System.Text;
	using System.Xml;
	using System.Runtime.InteropServices;

	using Microsoft.Win32;
	using Microsoft.Build.Framework;

	public class ComplusProcessor : Processor
	{
		/// <summary>
		/// Initializes the object.
		/// </summary>
		/// <param name="file">The <see cref="ITaskItem"/> file.</param>
		/// <param name="dirId">Sets the parent directories Id.</param>
		public ComplusProcessor ( ITaskItem file, string dirId ) : base ( file, dirId ) { }

		/// <summary>
		/// Gets the element prefix for the namespace.
		/// </summary>
		public override string ElementPrefix
		{
			get { return "pca"; }
		}

		/// <summary>
		/// Gets the namespace.
		/// </summary>
		public override string Namespace
		{
			get { return "http://schemas.microsoft.com/wix/2005/02/pubca"; }
		}
		/// <summary>
		/// Executes processing on the file.
		/// </summary>
		/// <param name="writer">The <see cref="XmlWriter"/> to append to.</param>
		public override void Execute ( XmlWriter writer )
		{
			IntPtr dllHandle = Common.LoadFile ( new System.IO.FileInfo ( File.GetMetadata ( "FullPath" ) ) );

			if ( dllHandle == IntPtr.Zero )
			{
				string err = string.Format ( "Error #{0}: {1}", Marshal.GetLastWin32Error ( ).ToString ( ),
					Common.GetErrorMessage ( Marshal.GetLastWin32Error ( ) ) );

				throw ( new System.IO.FileLoadException ( err, File.ItemSpec ) );
			}
			else
			{
				this.WriteFile ( writer );

				this.WriteComPlusAssembly ( writer );
			}
		}

		/// <summary>
		/// Writes the value of the parsed registry hives.
		/// </summary>
		/// <param name="writer">Optional writer to have override registry keys written to.</param>
		private void WriteRegistry ( XmlWriter writer )
		{
			if ( null != writer )
			{
				// read any generated registry keys
				using ( RegistryKey regKey = Registry.LocalMachine.OpenSubKey ( @"SOFTWARE\WICS\HKCR" ) )
				{
					string[] keys = regKey.GetSubKeyNames ( );

					foreach ( string key in keys )
					{
						string keyLower = key.ToLower ( );
						if ( !keyLower.Contains ( "clsid" ) || !keyLower.Contains ( "interface" ) || !keyLower.Contains ( "typelib" ) )
						{
							RegistryKey clsKey = regKey.OpenSubKey ( key );

							foreach ( string subKey in clsKey.GetSubKeyNames ( ) )
							{
								if ( subKey.ToLower ( ) == "clsid" )
								{
									RegistryKey clsid = clsKey.OpenSubKey ( subKey );
									writer.WriteStartElement ( "pca", "ComPlusComponent", null );
									writer.WriteAttributeString ( "Id", this.GetComPlusComponentId ( ) );
									writer.WriteAttributeString ( "CLSID", clsid.GetValue ( "" ).ToString ( ).Replace ( "{", "" ).Replace ( "}", "" ) );
									writer.WriteAttributeString ( "Description", key );
									writer.WriteEndElement ( );
								}
							}
						}
					}
				}
			}
		}

		private string GetComPlusAssemblyId ( )
		{
			string str = "ComPlusAssemblyId";
			if ( !string.IsNullOrEmpty ( File.GetMetadata ( str ) ) )
				return File.GetMetadata ( str );
			else
				return Common.Guids.CreateGuid ( "ComPlusAssemblyId", GuidManager.GuidType.Id );
		}

		private string GetComPlusAssemblyApplication ( )
		{
			string str = "ComPlusAssemblyApplication";

			if ( !string.IsNullOrEmpty ( this.File.GetMetadata ( str ) ) )
				return this.File.GetMetadata ( str );
			else if ( Common.Guids.ContainsKey ( str ) )
				return Common.Guids[str][0];
			else
			{
				Common.Guids.CreateGuid ( str, GuidManager.GuidType.Id );
				return Common.Guids[str][0];
			}
		}

		public string GetComPlusAssemblyType ( )
		{
			string str = "ComPlusAssemblyType";

			if ( !string.IsNullOrEmpty ( File.GetMetadata ( str ) ) )
				return File.GetMetadata ( str );
			else
				return "native";
		}

		/// <summary>
		/// Gets the ComPlusComponent Id from the metadata or creates a GUID if it isn't found.
		/// </summary>
		/// <returns>Returns a Id.</returns>
		private string GetComPlusComponentId ( )
		{
			string str = "ComPlusComponentId";
			if ( !string.IsNullOrEmpty ( File.GetMetadata ( str ) ) )
				return File.GetMetadata ( str );
			else
				return Common.Guids.CreateGuid ( "ComPlusComponentId", GuidManager.GuidType.Id );
		}

		/// <summary>
		/// Writes the COM+ assembly in WiX format.
		/// </summary>
		/// <param name="writer">A <see cref="XmlWriter"/> to append data.<</param>
		/// <param name="fileId">The Id of the file.</param>
		private void WriteComPlusAssembly ( XmlWriter writer )
		{
			writer.WriteStartElement ( "pca", "ComPlusAssembly", null );
			writer.WriteAttributeString ( "Id", this.GetComPlusAssemblyId ( ) );
			writer.WriteAttributeString ( "DllPath", "[#" + this.FileId + "]" );
			writer.WriteAttributeString ( "Application", this.GetComPlusAssemblyApplication ( ) );
			writer.WriteAttributeString ( "Type", this.GetComPlusAssemblyType ( ) );

			this.WriteComPlusComponent ( writer );

			writer.WriteEndElement ( );
		}

		/// <summary>
		/// Writes the COM+ component in WiX format.
		/// </summary>
		/// <param name="writer">A <see cref="XmlWriter"/> to append data.</param>
		private void WriteComPlusComponent ( XmlWriter writer )
		{
			bool success = false;

			try
			{
				Common.StartOverridingRegHives ( );

				Common.DynamicPInvoke ( File.GetMetadata ( "FullPath" ), "DllRegisterServer", typeof ( int ), null, null );

				success = true;
			}
			finally
			{
				try
				{
					Common.EndOverridingRegHives ( );
					this.WriteRegistry ( success ? writer : null );
				}
				finally
				{
					Common.RemovePrivateRegKey ( );	
				}
			}
		}
	}
}