namespace WICS.Generator.Processors
{
	using System;
	using System.Collections.Generic;
	using System.Text;
	using System.IO;

	using Microsoft.Build.Framework;

	/// <summary>
	/// This is use to manage all files that are to be processed. The object keeps a store
	/// of all files that have been found on the system and files that are not to be
	/// processed. When object is executed it will analize all files passed in the form
	/// of an array of <see cref="ITaskItem"/>. It will store all files that will not be
	/// processed in a <see cref="List"/> of <see cref="FileInfo"/> objects.
	/// </summary>
	public class FileManager : Dictionary<string, ITaskItem>
	{
		private List<ITaskItem> filesNotProcessed;
		private ITaskItem[] filesToProcess;
		private ITaskItem[] searchScope;
		private DirectoryInfo directory;
		private Dictionary<string, IProcessor> fileProcessors;
		private List<string> directories;

		/// <summary>
		/// Initializes the object.
		/// </summary>
		/// <param name="searchScope">An array <see cref="ITaskItem"/> that defines the search scope.</param>
		/// <param name="root">Holds the root directory that is being analyzed.</param>
		public FileManager ( ITaskItem[] searchScope, DirectoryInfo root ) : base ( StringComparer.OrdinalIgnoreCase )
		{
			this.filesNotProcessed = new List<ITaskItem> ( );
			this.searchScope = searchScope;
			this.fileProcessors = new Dictionary<string, IProcessor> ( );
			this.directories = new List<string> ( );
			this.directory = root;
		}

		/// <summary>
		/// Gets the root directory that is being analyzed.
		/// </summary>
		public DirectoryInfo Directory
		{
			get { return this.directory; }
		}

		/// <summary>
		/// Gets a list of directories that are part of the search scope.
		/// </summary>
		public List<string> Directories
		{
			get { return this.directories; }
		}

		/// <summary>
		/// Gets a list of files that will not be processed. The list is populated after the Analize method is called.
		/// </summary>
		public List<ITaskItem> FilesNotProcessed
		{
			get { return this.filesNotProcessed; }
		}

		/// <summary>
		/// Gets the array of <see cref="ITaskItem"/> files to process.
		/// </summary>
		public ITaskItem[] FilesToProcess
		{
			get { return this.filesToProcess; }
		}

		/// <summary>
		/// Gets a list of <see cref="IProcessor"/> types that have been found during analysis.
		/// </summary>
		public Dictionary<string, IProcessor> FileProcessors
		{
			get { return this.fileProcessors; }
		}

		/// <summary>
		/// Gets the search scope of file manager.
		/// </summary>
		public ITaskItem[] SearchScope
		{
			get { return this.searchScope; }
		}

		/// <summary>
		/// Runs an analysis of the files against the root directory.
		/// </summary>
		/// <param name="files">An <see cref="ITaskItem"/> array of files to process.</param>
		/// <exception cref="ArgumentOutOfRangeException">The file has already been added to
		/// the collection. Please check the list of files.</exception>
		public void Load ( ITaskItem[] files )
		{
			this.filesToProcess = files;

			this.AddFilesToProcess ( );

			this.FindNonProcessedFiles ( );
		}

		/// <summary>
		/// Add the correct name spaces and prefixes.
		/// </summary>
		/// <param name="writer">A <see cref="System.Xml.XmlWriter"/> to append to.</param>
		public void AddXmlDocumentNamespace ( System.Xml.XmlWriter writer )
		{
			foreach ( KeyValuePair<string, IProcessor> proc in this.FileProcessors )
			{
				if((proc.Value.ElementPrefix != null) && (proc.Value.Namespace != null))
					writer.WriteAttributeString ( "xmlns", proc.Value.ElementPrefix, null, proc.Value.Namespace );
			}
		}

		private void AddFilesToProcess ( )
		{
			foreach ( ITaskItem item in this.filesToProcess )
			{
				if ( File.Exists ( item.GetMetadata ( "FullPath" ) ) )
				{
					this.AddFileProcessor ( item );

					if ( !this.ContainsKey ( item.GetMetadata ( "FullPath" ) ) )
						this.Add ( item.GetMetadata ( "FullPath" ), item );
					else
						throw ( new ArgumentOutOfRangeException ( item.GetMetadata ( "FullPath" ),
							"The file has already been added to the collection. Please check the list of files and ItemGroups in your project." ) );
				}
				else
				{
					if ( !System.IO.Directory.Exists ( item.ItemSpec ) )
						throw ( new FileNotFoundException ( "The file ( " + item.ItemSpec + " ) was not found.", item.GetMetadata ( "FullPath" ) ) );
				}
			}
		}

		private void FindNonProcessedFiles ( )
		{
			foreach ( ITaskItem item in this.searchScope )
			{
				if ( !this.directories.Contains ( item.GetMetadata ( "RootDir" ) + item.GetMetadata ( "Directory" ) ) )
					this.directories.Add ( item.GetMetadata ( "RootDir" ) + item.GetMetadata ( "Directory" ) );

				if ( !this.ContainsKey ( item.GetMetadata ( "FullPath" ) ) && !System.IO.Directory.Exists ( item.GetMetadata ( "FullPath" ) ) )
					this.filesNotProcessed.Add ( item );
			}
		}

		private void AddFileProcessor ( ITaskItem item )
		{
			object[] args = new object[] { null, null };

			IProcessor processor = Activator.CreateInstance ( Type.GetType ( this.GetFileProcessor ( item ) ), args ) as IProcessor;
			
			if ( !this.FileProcessors.ContainsKey ( processor.GetType ( ).ToString ( ) ) )
				this.FileProcessors.Add ( processor.GetType ( ).ToString ( ), processor );
		}

		/// <summary>
		/// Gets the FileProcessor of the given item.
		/// </summary>
		/// <param name="item">A <see cref="ITaskItem"/> file.</param>
		/// <returns>A string value of the full processor namespace to use.
		/// <example>WICS.Generator.Processors.StaticProcessor</example>
		/// </returns>
		private string GetFileProcessor ( ITaskItem item )
		{
			string str = item.GetMetadata ( "FileProcessor" );

			if ( !string.IsNullOrEmpty ( str ) )
				return str;
			else
				return "WICS.Generator.Processors.StaticProcessor";
		}
	}
}