//-------------------------------------------------------------------------------------------------
// <copyright file="NativeMethods.cs" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//
//    You must not remove this notice, or any other, from this software.
// </copyright>
// 
// <summary>
// The tallow codegen tool application's Native Methods class.
// </summary>
//-------------------------------------------------------------------------------------------------
namespace WICS.Generator.Processors
{
	using System;
	using System.Collections.Generic;
	using System.Text;
	using System.Runtime.InteropServices;

	using Microsoft.Win32;

	/// <summary>
	/// Private class to do interop.
	/// </summary>
	internal class NativeMethods
	{
		private const int MaxPath = 255;

		private const uint StandardRightsAll = 0x001F0000;

		private const uint GenericRead = 0x80000000;
		private const uint GenericWrite = 0x40000000;
		private const uint GenericExecute = 0x20000000;
		private const uint GenericAll = 0x10000000;

		/// <summary>
		/// Gets the short name for a file.
		/// </summary>
		/// <param name="fullPath">Fullpath to file on disk.</param>
		/// <returns>Short name for file.</returns>
		internal static string GetShortPathName ( string fullPath )
		{
			StringBuilder shortPath = new StringBuilder ( MaxPath, MaxPath );

			// get the short file name
			GetShortPathName ( fullPath, shortPath, MaxPath );

			// remove the tildes
			shortPath.Replace ( '~', '_' );

			return shortPath.ToString ( );
		}

		/// <summary>
		/// Opens a registry key.
		/// </summary>
		/// <param name="key">Base key to open.</param>
		/// <param name="path">Path to subkey to open.</param>
		/// <returns>Handle to new key.</returns>
		internal static IntPtr OpenRegistryKey ( UIntPtr key, string path )
		{
			IntPtr newKey = IntPtr.Zero;
			uint disposition = 0;
			uint sam = StandardRightsAll | GenericRead | GenericWrite | GenericExecute | GenericAll;

			int er = RegCreateKeyEx ( key, path, 0, null, 0, sam, 0, out newKey, out disposition );

			return newKey;
		}

		/// <summary>
		/// Overrides a registry key.
		/// </summary>
		/// <param name="key">Handle to key to override.</param>
		/// <param name="newKey">Handle to override key.</param>
		internal static void OverrideRegistryKey ( UIntPtr key, IntPtr newKey )
		{
			int er = RegOverridePredefKey ( key, newKey );
		}

		/// <summary>
		/// Closes a previously open registry key.
		/// </summary>
		/// <param name="key">Handle to key to close.</param>
		internal static void CloseRegistryKey ( IntPtr key )
		{
			int er = RegCloseKey ( key );
		}

		/// <summary>
		/// Gets the short name for a file.
		/// </summary>
		/// <param name="longPath">Long path to convert to short path.</param>
		/// <param name="shortPath">Short path from long path.</param>
		/// <param name="buffer">Size of short path.</param>
		/// <returns>zero if success.</returns>
		[DllImport ( "kernel32.dll", EntryPoint = "GetShortPathNameW", CharSet = CharSet.Unicode, ExactSpelling = true, SetLastError = true )]
		internal static extern uint GetShortPathName ( string longPath, StringBuilder shortPath, [MarshalAs ( UnmanagedType.U4 )]int buffer );

		/// <summary>
		/// Interop to RegCreateKeyW
		/// </summary>
		/// <param name="key">Handle to base key.</param>
		/// <param name="subkey">Subkey to create.</param>
		/// <param name="reserved">Always 0</param>
		/// <param name="className">Just pass null.</param>
		/// <param name="options">Just pass 0.</param>
		/// <param name="desiredSam">Rights to registry key.</param>
		/// <param name="securityAttributes">Just pass null.</param>
		/// <param name="openedKey">Opened key.</param>
		/// <param name="disposition">Whether key was opened or created.</param>
		/// <returns>Handle to registry key.</returns>
		[DllImport ( "advapi32.dll", EntryPoint = "RegCreateKeyExW", CharSet = CharSet.Unicode, ExactSpelling = true, SetLastError = true )]
		internal static extern int RegCreateKeyEx ( UIntPtr key, string subkey, uint reserved, string className, uint options, uint desiredSam, uint securityAttributes, out IntPtr openedKey, out uint disposition );

		/// <summary>
		/// Interop to RegOverridePredefKey
		/// </summary>
		/// <param name="key">Handle to key to override.</param>
		/// <param name="newKey">Handle to override key.</param>
		/// <returns>0 if success.</returns>
		[DllImport ( "advapi32.dll", EntryPoint = "RegOverridePredefKey", CharSet = CharSet.Unicode, ExactSpelling = true, SetLastError = true )]
		internal static extern int RegOverridePredefKey ( UIntPtr key, IntPtr newKey );

		/// <summary>
		/// Interop to RegCloseKey
		/// </summary>
		/// <param name="key">Handle to key to close.</param>
		/// <returns>0 if success.</returns>
		[DllImport ( "advapi32.dll", EntryPoint = "RegCloseKey", CharSet = CharSet.Unicode, ExactSpelling = true, SetLastError = true )]
		internal static extern int RegCloseKey ( IntPtr key );
	}
}