namespace WICS.Generator.Processors
{
	using System;
	using System.Collections.Generic;
	using System.Text;
	using System.Xml;

	using Microsoft.Build.Framework;

	/// <summary>
	/// An base processor class that can be used to create additional processors.
	/// </summary>
	public abstract class Processor : IProcessor
	{
		private ITaskItem file;
		private string fileId;
		private string dirId;

		/// <summary>
		/// Initializes the object.
		/// </summary>
		/// <param name="file">The file to process as an <see cref="ITaskItem"/>.</param>
		/// <param name="dirId">Sets the parent directories Id.</param>
		public Processor ( ITaskItem file, string dirId )
		{
			this.file = file;
			this.dirId = dirId;
			this.GenerateFileId ( );
		}

		/// <summary>
		/// Gets the element prefix that will be added to the name.
		/// </summary>
		public virtual string ElementPrefix
		{
			get { return null; }
		}

		/// <summary>
		/// Gets the directory Id of the parent directory.
		/// </summary>
		public string DirectoryId
		{
			get { return this.dirId; }
		}

		/// <summary>
		/// Gets or sets the <see cref="ITaskItem"/> that represents the file to process.
		/// </summary>
		public ITaskItem File
		{
			get { return this.file; }
		}

		/// <summary>
		/// Gets the FileId.
		/// </summary>
		public string FileId
		{
			get { return this.fileId; }
		}

		/// <summary>
		/// Gets the namespace that will be added to the <see cref="XmlDocument"/>.
		/// </summary>
		public virtual string Namespace
		{
			get { return null; }
		}

		/// <summary>
		/// Executes the given object.
		/// </summary>
		/// <param name="writer">An <see cref="XmlWriter"/> to append to.</param>
		/// <returns>Returns true if the action was successful.</returns>
		public abstract void Execute ( XmlWriter writer );

		/// <summary>
		/// Generates a FileId for the processor and sets the property FileId with its value.
		/// </summary>
		private void GenerateFileId ( )
		{
			if ( this.File != null )
			{
				this.fileId = this.GetFileId ( );
			}
		}

		private string GetFileId ( )
		{
			string str = this.file.GetMetadata ( "FileId" );

			if ( !string.IsNullOrEmpty ( str ) )
				return str;
			else
				return Common.Guids.CreateGuid ( "FileId", GuidManager.GuidType.Id );
		}

		/// <summary>
		/// Writes the file to the given <see cref="XmlWriter"/>.
		/// </summary>
		/// <example>
		/// <code>
		///	System.IO.FileInfo fiFile = new System.IO.FileInfo ( File.ItemSpec );
		///	string shortName = Common.GenerateShortName ( fiFile );
		///	writer.WriteAttributeString ( "Name", shortName );
		///	if ( shortName != fiFile.Name )
		///		writer.WriteAttributeString ( "LongName", fiFile.Name );
		///
		///	writer.WriteAttributeString ( "Source", File.GetMetadata ( "FullPath" ) );
		/// </code>
		/// </example>
		/// <param name="writer">A <see cref="XmlWriter"/> to append to.</param>
		/// <exception cref="ArgumentNullException ">The file passed must not be null.</exception>
		/// <exception cref="System.IO.FileNotFoundException">The given file was not found.</exception>
		protected virtual void WriteFile ( XmlWriter writer )
		{
			if ( this.File == null )
				throw ( new ArgumentNullException ( "file", "The file passed must not be null." ) );

			if ( !System.IO.File.Exists ( file.ItemSpec ) )
				throw ( new System.IO.FileNotFoundException ( "The given file was not found.", file.ItemSpec ) );

			writer.WriteStartElement ( "File" );
			writer.WriteAttributeString ( "Id", this.FileId );
			writer.WriteAttributeString ( "KeyPath", "yes" );
			writer.WriteAttributeString ( "Vital", "yes" );

			System.IO.FileInfo fiFile = new System.IO.FileInfo ( File.ItemSpec );
			string shortName = Common.GenerateShortName ( fiFile );
			writer.WriteAttributeString ( "Name", shortName );
			if ( shortName != fiFile.Name )
				writer.WriteAttributeString ( "LongName", fiFile.Name );

			writer.WriteAttributeString ( "Source", File.GetMetadata ( "FullPath" ) );

			writer.WriteEndElement ( );
		}
	}
}